import logging
import re
from collections import defaultdict
from dataclasses import _MISSING_TYPE, dataclass, field, is_dataclass
from os import environ
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, get_type_hints

import pytz
import yaml
from yaml.loader import SafeLoader

from hartware_lib.types import Dataclass, T
from hartware_lib.utils.casing import pascal_to_snake_case

logger = logging.getLogger("hartware_lib.settings")


def transform(target_type: Type[T], value: Any) -> Optional[T]:
    if value is None:
        return value

    if target_type is bool:
        if not isinstance(value, str):
            value = str(value)
        if value.lower() in ("0", "no", "false"):
            return False  # type: ignore[return-value]
        if value.lower() in ("1", "yes", "true"):
            return True  # type: ignore[return-value]

        raise Exception(f"Can't parse '{value}' as boolean")

    return target_type(value)  # type: ignore[call-arg]


def load_settings(  # noqa: C901
    cls: Type[Dataclass],
    config: Optional[Dict[str, Any]] = None,
    path: Optional[List[str]] = None,
    set_prefix: str = "",
    hide_prefix: bool = False,
) -> Dataclass:
    settings = {}

    if path is None:
        path = []

    if config is None:
        config = {}

    if not hide_prefix:
        if set_prefix:
            prefix = set_prefix
        else:
            prefix = getattr(
                cls,
                "_prefix",
                pascal_to_snake_case(cls.__name__.replace("Settings", "")),
            )

        path = path + [prefix]

    type_hints = get_type_hints(cls)

    for option_name, option in cls.__dataclass_fields__.items():  # type: ignore[attr-defined]
        option_type = type_hints[option_name]

        option_name_target = option.metadata.get("prefix", option_name)

        env_path = "_".join(path + [option_name_target]).upper()
        config_value = config.get(option_name)

        if is_dataclass(option_type):
            if option.metadata.get("as_list"):
                config_values = {
                    i: load_settings(
                        option_type,
                        _config_value,
                        path + [option_name_target, str(i)],
                        hide_prefix=True,
                    )
                    for i, _config_value in enumerate(config_value or [])
                }

                extra_conf: Dict[int, Dict[Any, Any]] = defaultdict(dict)
                for k, v in environ.items():
                    if env_path not in k:
                        continue

                    m = re.match(rf"{env_path}_(?P<index>\d+)_(?P<attr>\w+)", k)
                    if not m:
                        logger.warning(f"Could not parse key: {k}")
                        continue
                    index = int(m.group("index"))
                    if index in config_values:
                        continue
                    extra_conf[index][m.group("attr")] = v
                for i, config_value in extra_conf.items():
                    try:
                        config_values[i] = load_settings(
                            option_type,
                            config_value,
                            path + [option_name_target, str(i)],
                            hide_prefix=True,
                        )
                    except Exception as exc:
                        logger.warning(f"Incomplete settings: {exc}")

                config_value = [
                    i[1] for i in sorted(config_values.items(), key=lambda x: x[0])
                ]
            else:
                config_value = load_settings(
                    option_type,
                    config_value,
                    path,
                    set_prefix=option.metadata.get("prefix", ""),
                )
        else:
            if env_value := environ.get(env_path):
                config_value = env_value
            if (cast_func := option.metadata.get("cast")) and isinstance(
                config_value, str
            ):
                config_value = cast_func(config_value)
            elif not isinstance(config_value, option_type):
                config_value = transform(option_type, config_value)
        if config_value is None:
            if not isinstance(option.default, _MISSING_TYPE):
                config_value = option.default
            elif not isinstance(option.default_factory, _MISSING_TYPE):
                config_value = option.default_factory()

            if not isinstance(config_value, option_type):
                raise Exception(f"No value for {'.'.join(path + [option_name])}")

        settings[option_name] = config_value

    return cls(**settings)  # type: ignore[call-arg]


def load_settings_from_file(
    base_dataclass: Type[Dataclass],
    config_path: Path,
    path: Optional[List[str]] = None,
) -> Dataclass:
    with open(config_path) as f:
        config = yaml.load(f, Loader=SafeLoader)

        return load_settings(base_dataclass, config or {}, path=path)


@dataclass
class RabbitMQSettings:
    _prefix = "rabbitmq"

    username: str
    password: str
    host: str
    port: int = 5672


@dataclass
class HttpRpcSettings:
    host: str
    port: int


@dataclass
class MetaTraderSettings:
    _prefix = "metatrader"

    login: int
    password: str
    path: Path
    server: str
    timezone: pytz.tzinfo.StaticTzInfo = field(metadata={"cast": pytz.timezone})


@dataclass
class SlackBotSettings:
    api_token: str
    default_channel: str



from __future__ import annotations

from functools import partial
from typing import Set

from hartware_lib.serializers.dataclasses import (
    dataclasses_extra_deserializer,
    dataclasses_extra_serializer,
)
from hartware_lib.serializers.main import deserialize, serialize
from hartware_lib.serializers.pandas import (
    pandas_extra_deserializer,
    pandas_extra_serializer,
)
from hartware_lib.types import (
    Dataclass,
    Deserializer,
    ExtraDeserializer,
    ExtraSerializer,
    Serializer,
)


class SerializerBuilder:
    def __init__(self) -> None:
        self.extra_serializers: Set[ExtraSerializer] = set()

    def add_serializer(self, extra_serializer: ExtraSerializer) -> SerializerBuilder:
        self.extra_serializers.add(extra_serializer)

        return self

    def with_pandas(self) -> SerializerBuilder:
        return self.add_serializer(pandas_extra_serializer)

    def with_dataclasses(self) -> SerializerBuilder:
        return self.add_serializer(dataclasses_extra_serializer)

    def get(self) -> Serializer:
        return partial(serialize, extra_serializers=list(self.extra_serializers))


class DeserializerBuilder:
    def __init__(self) -> None:
        self.extra_deserializers: Set[ExtraDeserializer] = set()

    def add_deserializer(
        self, extra_deserializer: ExtraDeserializer
    ) -> DeserializerBuilder:
        self.extra_deserializers.add(extra_deserializer)

        return self

    def with_pandas(self) -> DeserializerBuilder:
        return self.add_deserializer(pandas_extra_deserializer)

    def with_dataclasses(self, *dataclasses: Dataclass) -> DeserializerBuilder:
        return self.add_deserializer(dataclasses_extra_deserializer(*dataclasses))

    def get(self) -> Deserializer:
        return partial(deserialize, extra_deserializers=list(self.extra_deserializers))



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from typing import Any

from aiohttp import ClientSession, ClientTimeout

from hartware_lib.serializers.builders import DeserializerBuilder, SerializerBuilder
from hartware_lib.settings import HttpRpcSettings
from hartware_lib.types import AnyDict, Deserializer, Serializer

logger = logging.getLogger("hartware_lib.http_rpc_caller")


@dataclass
class HttpRpcCaller:
    settings: HttpRpcSettings
    serializer: Serializer
    deserializer: Deserializer

    @classmethod
    def build(
        cls,
        settings: HttpRpcSettings,
        serializer: Serializer = SerializerBuilder().get(),
        deserializer: Deserializer = DeserializerBuilder().get(),
    ) -> HttpRpcCaller:
        return cls(settings, serializer, deserializer)

    async def _process(self, data: AnyDict, timeout: float = 300.0) -> Any:
        async with ClientSession(
            timeout=ClientTimeout(timeout), raise_for_status=True
        ) as session:
            response = await session.post(
                f"http://{self.settings.host}:{self.settings.port}/",
                data={"order": self.serializer(data).decode("utf-8")},
                raise_for_status=True,
            )

            text = await response.text()

        data = self.deserializer(text)
        error = data.get("error")

        logger.info(f"received {len(text)} bytes ({type(data).__name__})")

        if error:
            raise Exception(f"{error}")

        return data.get("result")

    async def ping(self, timeout: float = 5.0) -> bool:
        logger.info("ping")

        try:
            result = await self._process({"ping": True}, timeout=timeout)

            if result.get("pong") is True:
                logger.info("pong received")

                return True
        except asyncio.exceptions.TimeoutError:
            logger.info("No pong received")
        except Exception as exc:
            logger.warning(f"No pong received: {exc}", exc_info=True)

        return False

    async def get_property(self, name: str, timeout: float = 10.0) -> Any:
        logger.info(f"get_property: {name}")

        return await self._process({"property": name}, timeout=timeout)

    async def set_property(self, name: str, value: Any, timeout: float = 10.0) -> None:
        logger.info(f"set_property: {name} to {value:r}")

        await self._process({"property": name, "property_set": value}, timeout=timeout)

    async def call(
        self, func: str, *args: Any, timeout: float = 300.0, **kwargs: Any
    ) -> Any:
        logger.info(f"call: {str(func)} = *{args}, **{kwargs}")

        return await self._process(
            {"func": func, "args": args, "kwargs": kwargs}, timeout=timeout
        )



from datetime import datetime, timedelta


def get_next_datetime(now: datetime, interval: timedelta) -> datetime:
    now_in_seconds = (now - datetime.min).total_seconds()
    interval_in_seconds = interval.total_seconds()

    quotient = (now_in_seconds // interval_in_seconds) + 1

    return datetime.min + timedelta(seconds=quotient * interval_in_seconds)



from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Any, AsyncIterator

from hartware_lib.adapters.rabbitmq import (
    RabbitMQAdapter,
    RabbitMQDefaultExchangeAdapter,
)
from hartware_lib.controllers.rmq_rpc_caller import RpcCaller
from hartware_lib.controllers.rmq_rpc_probe import RpcProbe
from hartware_lib.settings import RabbitMQSettings


@dataclass
class RmqRpcController:
    rpc_in: RabbitMQDefaultExchangeAdapter
    rpc_out: RabbitMQDefaultExchangeAdapter

    @classmethod
    def build(
        cls, rabbitmq_settings: RabbitMQSettings, queue_name: str, **kwargs: Any
    ) -> RmqRpcController:
        rabbitmq = RabbitMQAdapter.build(settings=rabbitmq_settings, **kwargs)

        rpc_in = rabbitmq.get_flavor_adapter("default", routing_key=f"{queue_name}_in")
        rpc_out = rabbitmq.get_flavor_adapter(
            "default", routing_key=f"{queue_name}_out"
        )

        assert isinstance(rpc_in, RabbitMQDefaultExchangeAdapter)
        assert isinstance(rpc_out, RabbitMQDefaultExchangeAdapter)

        return cls(rpc_in, rpc_out)

    @asynccontextmanager
    async def connected(self) -> AsyncIterator[None]:
        async with await self.rpc_in.connect():
            async with await self.rpc_out.connect():
                yield

    def get_probe(self, subject: object) -> RpcProbe:
        return RpcProbe(self.rpc_in, self.rpc_out, subject)

    def get_caller(self) -> RpcCaller:
        return RpcCaller(self.rpc_in, self.rpc_out)



from typing import Any, Callable, Dict, TypeVar, Union

T = TypeVar("T")

HeadersDict = Dict[str, str]
AnyDict = Dict[str, Any]
Dataclass = TypeVar("Dataclass")

Serializer = Callable[[Any], bytes]
Deserializer = Callable[[Union[bytes, str]], Any]

ExtraSerializer = Callable[[Any], AnyDict]
ExtraDeserializer = Callable[[AnyDict], Any]



from __future__ import annotations

import asyncio
import logging
import traceback
from dataclasses import dataclass
from typing import Any

from aiohttp import web

from hartware_lib.serializers.builders import DeserializerBuilder, SerializerBuilder
from hartware_lib.settings import HttpRpcSettings
from hartware_lib.types import Deserializer, Serializer

logger = logging.getLogger("hartware_lib.http_rpc_probe")


@dataclass
class HttpRpcProbe:
    app: web.Application
    runner: web.AppRunner
    subject: Any
    settings: HttpRpcSettings
    serializer: Serializer
    deserializer: Deserializer

    @classmethod
    def build(
        cls,
        settings: HttpRpcSettings,
        subject: Any,
        serializer: Serializer = SerializerBuilder().get(),
        deserializer: Deserializer = DeserializerBuilder().get(),
    ) -> HttpRpcProbe:
        app = web.Application()
        runner = web.AppRunner(app)

        obj = cls(app, runner, subject, settings, serializer, deserializer)

        app.add_routes([web.post("/", obj.handle)])

        return obj

    async def handle(self, request: web.Request) -> web.Response:
        data = (await request.post())["order"]
        assert isinstance(data, str)

        order = self.deserializer(data)
        assert isinstance(order, dict)

        if "ping" in order:
            logger.info("ping received, pong sent")

            return web.Response(body=self.serializer({"result": {"pong": True}}))

        func = order.get("func")
        property = order.get("property")
        property_set = order.get("property_set")
        args = order.get("args") or []
        kwargs = order.get("kwargs") or {}

        if not func and not property:
            return web.Response(
                body=self.serializer(
                    {"error": "should have func or property specified"}
                ),
            )

        result = None
        try:
            if func:
                logger.info(f"call: {str(func)} = {args=}, {kwargs=}")

                func = getattr(self.subject, func)

                if asyncio.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
            else:
                assert isinstance(property, str)

                if "property_set" in order:
                    logger.info(f"set_property: {property} to {property_set:r}")

                    setattr(self.subject, property, property_set)
                else:
                    logger.info(f"get_property: {property}")

                    result = getattr(self.subject, property)
        except Exception:
            logger.info("got an exception:", exc_info=True)

            return web.Response(body=self.serializer({"error": traceback.format_exc()}))

        body = self.serializer({"result": result})

        logger.info(f"returns {len(body)} bytes ({type(result).__name__})")

        return web.Response(body=body)

    async def run(self) -> None:
        logger.info("start")

        await self.runner.setup()

        site = web.TCPSite(self.runner, self.settings.host, self.settings.port)
        await site.start()

        await asyncio.Future()

    async def cleanup(self) -> None:
        logger.info("cleaning up")

        await self.runner.cleanup()

        logger.info("stopped")



from __future__ import annotations

import asyncio
from typing import Any, Dict, Optional, Union

from hartware_lib.adapters.rabbitmq import RabbitMQDefaultExchangeAdapter


class RpcCaller:
    _none_value = object()
    _max_wait_loop = 150

    def __init__(
        self,
        rpc_in: RabbitMQDefaultExchangeAdapter,
        rpc_out: RabbitMQDefaultExchangeAdapter,
    ):
        self.rpc_in = rpc_in
        self.rpc_out = rpc_out

        self._result: Union[object, Dict[str, Any]] = self._none_value
        self.command_lock = asyncio.Lock()
        self.consumer_task: Optional[asyncio.Task[Any]] = None

    def listen(self) -> None:
        self.consumer_task = asyncio.create_task(self.rpc_out.consume(self._store_result))  # type: ignore[arg-type]

    async def _store_result(
        self, _origin: RabbitMQDefaultExchangeAdapter, result: Dict[str, Any]
    ) -> None:
        self._result = result

    async def get_property(self, name: str) -> Any:
        return await self._process({"property": name})

    async def set_property(self, name: str, value: Any) -> None:
        await self._process({"property": name, "property_set": value})

    async def call(self, func: str, *args: Any, **kwargs: Any) -> Any:
        return await self._process({"func": func, "args": args, "kwargs": kwargs})

    async def _process(self, request: Dict[str, Any]) -> Any:
        self._result = self._none_value

        async with self.command_lock:
            await self.rpc_in.publish(request)

            for _ in range(self._max_wait_loop):
                if self._result != self._none_value:
                    assert isinstance(self._result, dict)

                    error = self._result.get("error")

                    if error:
                        raise Exception(f"{error}")

                    return self._result.get("result")
                await asyncio.sleep(0.2)
            raise Exception("Command Timeout")

    async def stop(self) -> None:
        if self.consumer_task and not self.consumer_task.done():
            self.consumer_task.cancel()

            await asyncio.wait([self.consumer_task])



from typing import Any, Optional

from aiohttp import ClientResponse, ClientSession

from hartware_lib.types import AnyDict, HeadersDict


class HttpAsyncAdapter:
    def __init__(self, headers: Optional[HeadersDict] = None):
        self.headers = headers or {}

    async def request(
        self,
        method: str,
        url: str,
        extra_headers: Optional[HeadersDict] = None,
        **kwargs: Any
    ) -> ClientResponse:
        headers = self.headers

        if extra_headers:
            headers.update(extra_headers)

        async with ClientSession() as session:
            response = await session.request(method, url, headers=headers, **kwargs)
            response.raise_for_status()

            return response

    async def get_text(self, *args: Any, **kwargs: Any) -> str:
        response = await self.request(*args, **kwargs)

        return await response.text()

    async def get_json(self, *args: Any, **kwargs: Any) -> AnyDict:
        response = await self.request(*args, **kwargs)

        return await response.json()  # type: ignore[no-any-return]



from dataclasses import is_dataclass
from typing import Any, Callable

from hartware_lib.serializers.main import NoSerializerMatch
from hartware_lib.types import AnyDict, Dataclass


def dataclasses_extra_serializer(o: Any) -> AnyDict:
    if is_dataclass(o):
        return {"_type": o.__class__.__name__, "value": o.__dict__}

    raise NoSerializerMatch()


def dataclasses_extra_deserializer(*dataclasses: Dataclass) -> Callable[[AnyDict], Any]:
    registered_dataclasses = {
        dataclass.__name__: dataclass for dataclass in dataclasses  # type: ignore[attr-defined]
    }

    def wrapper(obj: AnyDict) -> Any:
        if obj_type := obj.get("_type"):
            obj_value = obj["value"]

            if dataclass := registered_dataclasses.get(obj_type):
                return dataclass(**obj_value)  # type: ignore[operator]

        raise NoSerializerMatch()

    return wrapper



from __future__ import annotations

import logging.config
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict


@dataclass
class LoggingBuilder:
    config: Dict[str, Any] = field(
        default_factory=lambda: {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "raw": {"format": "%(message)s"},
                "standard": {
                    "format": "%(asctime)s %(levelname)-8s %(name)-22s: %(message)s"
                },
            },
            "handlers": {},
            "loggers": {
                "": {"handlers": [], "level": "DEBUG", "propagate": False},
            },
        }
    )

    def add_default_handler(
        self, level: int = logging.DEBUG, formatter: str = "standard"
    ) -> LoggingBuilder:
        self.config["handlers"]["default"] = {
            "level": level,
            "formatter": formatter,
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        }
        self.config["loggers"][""]["handlers"].append("default")

        return self

    def add_file_handler(
        self,
        file_path: Path,
        level: int = logging.DEBUG,
        formatter: str = "standard",
    ) -> LoggingBuilder:
        self.config["handlers"]["file"] = {
            "level": level,
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": formatter,
            "filename": str(file_path),
            "maxBytes": 10_000_000,
            "backupCount": 10,
        }
        self.config["loggers"][""]["handlers"].append("file")

        return self

    def add_logger(
        self, name: str, level: int = logging.DEBUG, propagate: bool = False
    ) -> LoggingBuilder:
        root_handlers = self.config["loggers"][""]["handlers"]

        self.config["loggers"][name] = {
            "handlers": root_handlers,
            "level": level,
            "propagate": propagate,
        }

        return self

    def set_formatter(self, name: str, format: str) -> LoggingBuilder:
        self.config["formatters"][name] = {"format": format}

        return self

    def apply(self, func: Callable[[LoggingBuilder], None]) -> LoggingBuilder:
        func(self)

        return self

    def shut(self, module_name: str, level: int = logging.WARNING) -> LoggingBuilder:
        logging.getLogger(module_name).setLevel(level)

        return self

    def run(self) -> None:
        logging.config.dictConfig(self.config)



from io import StringIO
from typing import Any

from pandas import DataFrame, Series, read_json

from hartware_lib.serializers.main import NoSerializerMatch
from hartware_lib.types import AnyDict


def pandas_extra_serializer(o: Any) -> AnyDict:
    if isinstance(o, DataFrame):
        return {
            "_type": o.__class__.__name__,
            "index": o.index.name,
            "value": o.reset_index().to_json(orient="records"),
        }
    if isinstance(o, Series):
        return {"_type": o.__class__.__name__, "value": o.to_json()}

    raise NoSerializerMatch()


def pandas_extra_deserializer(obj: AnyDict) -> Any:
    if obj_type := obj.get("_type"):
        obj_value = obj["value"]

        if obj_type == "DataFrame":
            df = read_json(StringIO(obj_value), orient="records")

            if obj["index"]:
                df = df.set_index(obj["index"])
            else:
                df = df.set_index("index")
                df.index.name = None

            return df
        if obj_type == "Series":
            return read_json(StringIO(obj_value), typ="series")

    raise NoSerializerMatch()



from hartware_lib.serializers.builders import DeserializerBuilder, SerializerBuilder
from hartware_lib.serializers.main import NoSerializerMatch, deserialize, serialize

__all__ = (
    "deserialize",
    "serialize",
    "DeserializerBuilder",
    "SerializerBuilder",
    "NoSerializerMatch",
)



from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from glob import glob
from pathlib import Path
from shutil import rmtree
from typing import IO, AsyncIterator, Iterator, List, Union

import aiofiles
from aiofiles.threadpool.text import AsyncTextIOWrapper


class FileAlreadyExists(Exception):
    pass


class DirectoryAlreadyExists(Exception):
    pass


@dataclass
class FileAdapter:
    path: Path

    def touch(self, raise_if_exists: bool = False) -> bool:
        if self.exists:
            if raise_if_exists:
                raise FileAlreadyExists()

            return False

        self.directory.create()

        return self.write("")

    @property
    def directory(self) -> DirectoryAdapter:
        return DirectoryAdapter(self.path.parent)

    @property
    def exists(self) -> bool:
        return self.path.exists()

    def delete(self) -> bool:
        if self.exists:
            self.path.unlink()

            return True

        return False

    @contextmanager
    def open(self, mode: str = "r") -> Iterator[IO[str]]:
        with open(self.path, mode=mode) as file:
            yield file

    def read(self) -> str:
        with self.open(mode="r") as file:
            return file.read()

    def write(self, data: str = "") -> bool:
        with self.open(mode="w") as file:
            file.write(data)

            return True

    @asynccontextmanager
    async def async_open(self, mode: str = "r") -> AsyncIterator[AsyncTextIOWrapper]:
        async with aiofiles.open(self.path, mode=mode) as file:  # type: ignore[call-overload]
            yield file

    async def async_read(self) -> str:
        async with self.async_open(mode="r") as file:
            return await file.read()

    async def async_write(self, data: str = "") -> bool:
        async with self.async_open(mode="w") as file:
            await file.write(data)
            await file.flush()

            return True


@dataclass
class DirectoryAdapter:
    path: Path

    @property
    def exists(self) -> bool:
        return self.path.exists()

    @property
    def parent(self) -> DirectoryAdapter:
        return DirectoryAdapter(self.path.parent)

    def create(self, raise_if_exists: bool = False) -> bool:
        if self.exists:
            if raise_if_exists:
                raise DirectoryAlreadyExists()

            return False

        self.path.mkdir(parents=True)

        return True

    def file(self, path: Union[str, Path]) -> FileAdapter:
        return FileAdapter(self.path / path)

    def delete(self) -> bool:
        if self.exists:
            rmtree(self.path)

            return True

        return False

    def sub(self, sub_dir_path: Path) -> DirectoryAdapter:
        return DirectoryAdapter(self.path / sub_dir_path)

    def search(self, path_regex: Union[str, Path] = "*") -> List[FileAdapter]:
        return [
            FileAdapter(Path(path))
            for path in glob(str(self.path / "**" / path_regex), recursive=True)
        ]



import json
from datetime import datetime
from decimal import Decimal
from functools import partial
from pathlib import Path
from typing import Any, Optional, Sequence, Union

from hartware_lib.types import AnyDict, ExtraDeserializer, ExtraSerializer


class NoSerializerMatch(Exception):
    pass


class GlobalEncoder(json.JSONEncoder):
    def __init__(
        self,
        *args: Any,
        extra_serializers: Sequence[ExtraSerializer] = (),
        **kwargs: Any,
    ):
        super(GlobalEncoder, self).__init__(*args, **kwargs)

        self.extra_serializers = extra_serializers

    def default(self, o: Any) -> AnyDict:
        if isinstance(o, datetime):
            return {"_type": o.__class__.__name__, "value": o.isoformat()}
        elif isinstance(o, set):
            return {"_type": o.__class__.__name__, "value": list(o)}
        elif isinstance(o, Decimal):
            return {"_type": o.__class__.__name__, "value": str(o)}
        elif isinstance(o, Path):
            return {"_type": "Path", "value": str(o)}
        elif self.extra_serializers:
            for extra_serializer in self.extra_serializers:
                try:
                    return extra_serializer(o)
                except NoSerializerMatch:
                    pass

        raise NoSerializerMatch(f"Unknown `{o.__class__.__name__}` type")


def serialize(
    obj: Any,
    indent: Optional[int] = None,
    extra_serializers: Sequence[ExtraSerializer] = (),
) -> bytes:
    return (
        GlobalEncoder(indent=indent, extra_serializers=extra_serializers)
        .encode(obj)
        .encode("utf-8")
    )


def _global_decoder(
    obj: AnyDict, extra_deserializers: Sequence[ExtraSerializer]
) -> Any:
    if obj_type := obj.get("_type"):
        obj_value = obj["value"]

        if obj_type == "datetime":
            return datetime.fromisoformat(obj_value)
        elif obj_type == "set":
            return set(obj_value)
        elif obj_type == "Decimal":
            return Decimal(obj_value)
        elif obj_type == "Path":
            return Path(obj_value)
        elif extra_deserializers:
            for extra_deserializer in extra_deserializers:
                try:
                    return extra_deserializer(obj)
                except NoSerializerMatch:
                    pass

        raise NoSerializerMatch(f"Unknown `{obj_type}` type")

    return obj


def deserialize(
    obj: Union[bytes, str], extra_deserializers: Sequence[ExtraDeserializer] = ()
) -> Any:
    return json.loads(
        obj,
        object_hook=partial(_global_decoder, extra_deserializers=extra_deserializers),
    )



from __future__ import annotations

import asyncio
import logging
from typing import Any, Awaitable, Callable, List, Optional

from aio_pika import (
    DeliveryMode,
    ExchangeType,
    Message,
    RobustConnection,
    connect_robust,
)
from aio_pika.abc import AbstractIncomingMessage

from hartware_lib.serializers import deserialize as default_deserializer
from hartware_lib.serializers import serialize as default_serializer
from hartware_lib.settings import RabbitMQSettings
from hartware_lib.types import Deserializer, Serializer

logger = logging.getLogger("hartware_lib.rabbitmq")


class RabbitMQAdapter:
    def __init__(
        self,
        settings: RabbitMQSettings,
        serializer: Serializer,
        deserializer: Deserializer,
        object_class: Any = object,
    ):
        self.settings = settings
        self.serializer = serializer
        self.deserializer = deserializer
        self.object_class = object_class

    @classmethod
    def build(
        cls,
        settings: RabbitMQSettings,
        serializer: Serializer = default_serializer,
        deserializer: Deserializer = default_deserializer,
        object_class: Any = dict,
    ) -> RabbitMQAdapter:
        return cls(settings, serializer, deserializer, object_class)

    async def get_connection(self) -> RobustConnection:
        return await connect_robust(  # type: ignore[return-value]
            f"amqp://{self.settings.username}:{self.settings.password}@{self.settings.host}:{self.settings.port}/"
        )

    async def wait_until_ready(self, n: int = 30, delay: float = 1.0) -> None:
        for i in range(n):
            try:
                connection = await self.get_connection()

                if not connection.is_closed:
                    await connection.close()

                    return None
            except OSError:
                if i == n - 1:
                    raise

                await asyncio.sleep(delay)

    def handle_message(self, callback: CallbackHandler) -> MessageHandler:
        async def wrapper(message: AbstractIncomingMessage) -> Any:
            async with message.process():
                obj = self.deserializer(message.body)

                return await callback(self, obj)

        return wrapper

    def encode_object(self, obj: Any) -> bytes:
        if not isinstance(obj, self.object_class):
            raise Exception(f"Message must be of type `{self.object_class.__name__}`")

        return self.serializer(obj)

    def to_message(self, obj: Any) -> Message:
        message_body = self.encode_object(obj)

        return Message(message_body, delivery_mode=DeliveryMode.PERSISTENT)

    async def run_message_on_default_exchange(
        self, connection: RobustConnection, obj: Any, routing_key: str
    ) -> None:
        channel = await connection.channel()

        await channel.default_exchange.publish(
            self.to_message(obj), routing_key=routing_key
        )

    async def run_consumer_on_default_exchange(
        self, connection: RobustConnection, callback: CallbackHandler, routing_key: str
    ) -> None:
        channel = await connection.channel()

        await channel.set_qos(prefetch_count=1)

        queue = await channel.declare_queue(routing_key, durable=True)

        await queue.consume(self.handle_message(callback))

        await asyncio.Future()

    async def run_message_on_fanout_exchange(
        self,
        connection: RobustConnection,
        obj: Any,
        exchange_name: str,
        routing_key: str = "none",
    ) -> None:
        channel = await connection.channel()
        exchange = await channel.declare_exchange(exchange_name, ExchangeType.FANOUT)

        await exchange.publish(self.to_message(obj), routing_key=routing_key)

    async def run_consumer_on_fanout_exchange(
        self,
        connection: RobustConnection,
        callback: CallbackHandler,
        exchange_name: str,
    ) -> None:
        channel = await connection.channel()

        await channel.set_qos(prefetch_count=1)

        exchange = await channel.declare_exchange(exchange_name, ExchangeType.FANOUT)
        queue = await channel.declare_queue(exclusive=True)

        await queue.bind(exchange)
        await queue.consume(self.handle_message(callback))

        await asyncio.Future()

    async def run_message_on_topic_exchange(
        self,
        connection: RobustConnection,
        obj: Any,
        exchange_name: str,
        routing_key: str,
    ) -> None:
        channel = await connection.channel()
        exchange = await channel.declare_exchange(exchange_name, ExchangeType.TOPIC)

        await exchange.publish(self.to_message(obj), routing_key=routing_key)

    async def run_consumer_on_topic_exchange(
        self,
        connection: RobustConnection,
        callback: CallbackHandler,
        exchange_name: str,
        binding_keys: List[str],
    ) -> None:
        channel = await connection.channel()

        await channel.set_qos(prefetch_count=1)

        exchange = await channel.declare_exchange(exchange_name, ExchangeType.TOPIC)
        queue = await channel.declare_queue("", durable=True)

        for binding_key in binding_keys:
            await queue.bind(exchange, routing_key=binding_key)

        await queue.consume(self.handle_message(callback))

        await asyncio.Future()

    def get_flavor_adapter(
        self, flavor: str, *args: Any, **kwargs: Any
    ) -> RabbitMQFlavor:
        kwargs.update(self.__dict__)

        if flavor == "default":
            return RabbitMQDefaultExchangeAdapter(*args, **kwargs)
        elif flavor == "fanout":
            return RabbitMQFanoutExchangeAdapter(*args, **kwargs)
        elif flavor == "topic":
            return RabbitMQTopicExchangeAdapter(*args, **kwargs)
        else:
            raise Exception(f"No flavor '{flavor}'")


CallbackHandler = Callable[[RabbitMQAdapter, Any], Awaitable[Any]]
MessageHandler = Callable[[AbstractIncomingMessage], Awaitable[Any]]


class RabbitMQFlavor(RabbitMQAdapter):
    def __init__(self, *args: Any, **kwargs: Any):
        super(RabbitMQFlavor, self).__init__(*args, **kwargs)

        self._connection: Optional[RobustConnection] = None

    async def ensure_connection(self) -> RobustConnection:
        if self._connection is None:
            self._connection = await self.get_connection()

        return self._connection

    async def connect(self) -> RobustConnection:
        return await self.ensure_connection()

    @property
    async def connected(self) -> bool:
        return bool(await self.ensure_connection())


class RabbitMQDefaultExchangeAdapter(RabbitMQFlavor):
    def __init__(self, *args: Any, routing_key: str, **kwargs: Any):
        super(RabbitMQDefaultExchangeAdapter, self).__init__(*args, **kwargs)

        self.routing_key = routing_key

    async def publish(self, obj: Any) -> None:
        connection = await self.ensure_connection()

        await self.run_message_on_default_exchange(connection, obj, self.routing_key)

    async def consume(self, callback: CallbackHandler) -> None:
        connection = await self.ensure_connection()

        await self.run_consumer_on_default_exchange(
            connection, callback, self.routing_key
        )


class RabbitMQFanoutExchangeAdapter(RabbitMQFlavor):
    def __init__(self, *args: Any, exchange_name: str, **kwargs: Any):
        super(RabbitMQFanoutExchangeAdapter, self).__init__(*args, **kwargs)

        self.exchange_name = exchange_name

    async def publish(self, obj: Any) -> None:
        connection = await self.ensure_connection()

        await self.run_message_on_fanout_exchange(connection, obj, self.exchange_name)

    async def consume(self, callback: CallbackHandler) -> None:
        connection = await self.ensure_connection()

        await self.run_consumer_on_fanout_exchange(
            connection, callback, self.exchange_name
        )


class RabbitMQTopicExchangeAdapter(RabbitMQFlavor):
    def __init__(self, *args: Any, exchange_name: str, **kwargs: Any):
        super(RabbitMQTopicExchangeAdapter, self).__init__(*args, **kwargs)

        self.exchange_name = exchange_name

    async def publish(self, obj: Any, routing_key: str) -> None:
        connection = await self.ensure_connection()

        await self.run_message_on_topic_exchange(
            connection, obj, self.exchange_name, routing_key
        )

    async def consume(
        self, callback: CallbackHandler, binding_keys: List[str], **kwargs: Any
    ) -> None:
        connection = await self.ensure_connection()

        await self.run_consumer_on_topic_exchange(
            connection, callback, self.exchange_name, binding_keys
        )



from __future__ import annotations

import asyncio
import traceback
from typing import Any

from hartware_lib.adapters.rabbitmq import RabbitMQDefaultExchangeAdapter


class RpcProbe:
    def __init__(
        self,
        rpc_in: RabbitMQDefaultExchangeAdapter,
        rpc_out: RabbitMQDefaultExchangeAdapter,
        subject: Any,
    ):
        self.rpc_in = rpc_in
        self.rpc_out = rpc_out
        self.subject = subject

    async def run(self) -> None:
        await self.rpc_in.consume(self.handle_message)  # type: ignore[arg-type]

    async def handle_message(
        self, _origin: RabbitMQDefaultExchangeAdapter, message: Any
    ) -> None:
        assert isinstance(message, dict)

        func = message.get("func")
        property = message.get("property")
        property_set = message.get("property_set")
        args = message.get("args") or []
        kwargs = message.get("kwargs") or {}

        if not func and not property:
            await self.rpc_out.publish(
                {"error": "should have func or property specified"}
            )

            return

        result = None
        try:
            if func:
                func = getattr(self.subject, func)

                if asyncio.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
            else:
                assert isinstance(property, str)

                if "property_set" in message:
                    setattr(self.subject, property, property_set)
                else:
                    result = getattr(self.subject, property)
        except Exception:
            await self.rpc_out.publish({"error": traceback.format_exc()})

            return

        await self.rpc_out.publish({"result": result})
